
/*
 * Copyright (C) Nginx, Inc.
 * Copyright (C) Valentin V. Bartenev
 * Copyright (C) Ruslan Ermilov
 */


#include <ngx_config.h>
#include <ngx_core.h>
#include <ngx_thread_pool.h>
#include "dlb.h"

//TODO, add to config
#define NUM_EVENTS_PER_BATCH 1
#define DLB_NUM  4

typedef struct {
    ngx_array_t               pools;
} ngx_thread_pool_conf_t;


typedef struct {
    ngx_thread_task_t        *first;
    ngx_thread_task_t       **last;
} ngx_thread_pool_queue_t;

#define ngx_thread_pool_queue_init(q)                                         \
    (q)->first = NULL;                                                        \
    (q)->last = &(q)->first


struct ngx_thread_pool_s {
    ngx_thread_mutex_t        mtx;
    ngx_thread_pool_queue_t   queue;
    ngx_int_t                 waiting;
    ngx_thread_cond_t         cond;

    ngx_log_t                *log;

    ngx_str_t                 name;
    ngx_uint_t                threads;
    ngx_int_t                 max_queue;

    u_char                   *file;
    ngx_uint_t                line;

    ngx_uint_t                loadbal;
    ngx_ldb_t                 loadbal_queue;
    ngx_thread_pool_queue_t   p_queue[8]; //sw priority queue
    ngx_atomic_uint_t         p_recieved[8];
    ngx_atomic_uint_t         p_send[8];
};

typedef struct {
  dlb_port_hdl_t     rx_port;
  ngx_thread_pool_t* tp;
} ngx_thread_param_t;


enum wait_mode_t {
    POLL,
    INTERRUPT,
} wait_mode = POLL;

static ngx_int_t ngx_thread_pool_init(ngx_thread_pool_t *tp, ngx_log_t *log,
    ngx_pool_t *pool);
static void ngx_thread_pool_destroy(ngx_thread_pool_t *tp);
static void ngx_thread_pool_exit_handler(void *data, ngx_log_t *log);

static void *ngx_thread_pool_cycle_loadbalance(void *data);
static void *ngx_thread_pool_cycle(void *data);
static void ngx_thread_pool_handler(ngx_event_t *ev);

static char *ngx_thread_pool(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);

static void *ngx_thread_pool_create_conf(ngx_cycle_t *cycle);
static char *ngx_thread_pool_init_conf(ngx_cycle_t *cycle, void *conf);

static ngx_int_t ngx_thread_pool_init_worker(ngx_cycle_t *cycle);
static void ngx_thread_pool_exit_worker(ngx_cycle_t *cycle);


static ngx_command_t  ngx_thread_pool_commands[] = {

    { ngx_string("thread_pool"),
      NGX_MAIN_CONF|NGX_DIRECT_CONF|NGX_CONF_TAKE23,
      ngx_thread_pool,
      0,
      0,
      NULL },

      ngx_null_command
};


static ngx_core_module_t  ngx_thread_pool_module_ctx = {
    ngx_string("thread_pool"),
    ngx_thread_pool_create_conf,
    ngx_thread_pool_init_conf
};


ngx_module_t  ngx_thread_pool_module = {
    NGX_MODULE_V1,
    &ngx_thread_pool_module_ctx,           /* module context */
    ngx_thread_pool_commands,              /* module directives */
    NGX_CORE_MODULE,                       /* module type */
    NULL,                                  /* init master */
    NULL,                                  /* init module */
    ngx_thread_pool_init_worker,           /* init process */
    NULL,                                  /* init thread */
    NULL,                                  /* exit thread */
    ngx_thread_pool_exit_worker,           /* exit process */
    NULL,                                  /* exit master */
    NGX_MODULE_V1_PADDING
};


static ngx_str_t  ngx_thread_pool_default = ngx_string("default");

static ngx_uint_t               ngx_thread_pool_task_id;
static ngx_atomic_t             ngx_thread_pool_done_lock;
static ngx_thread_pool_queue_t  ngx_thread_pool_done;

#if (AIO_LOADBALANCE)

#define CQ_DEPTH 2
static int partial_resources = 4;

static int create_sched_domain(
    dlb_hdl_t dlb,
    ngx_ldb_t *ldb_p)
{
    dlb_create_sched_domain_t args;
    int p_rsrsc = partial_resources;

    //TODO, based on config
    args.num_ldb_queues = 1;
    args.num_ldb_ports = 2;
    args.num_dir_ports = 2;
    args.num_ldb_event_state_entries = 2 * args.num_ldb_ports * CQ_DEPTH;

    //if (!cap.combined_credits) {
    if (1) {
        args.num_ldb_credits = ldb_p->rsrcs.max_contiguous_ldb_credits * p_rsrsc / 100;
        args.num_dir_credits = ldb_p->rsrcs.max_contiguous_dir_credits * p_rsrsc / 100;
        args.num_ldb_credit_pools = 1;
        args.num_dir_credit_pools = 1;
    } else {
        args.num_credits = ldb_p->rsrcs.num_credits * p_rsrsc / 100;
        args.num_credit_pools = 1;
    }

    args.num_sn_slots[0] = ldb_p->rsrcs.num_sn_slots[0] * p_rsrsc / 100;
    args.num_sn_slots[1] = ldb_p->rsrcs.num_sn_slots[1] * p_rsrsc / 100;

    return dlb_create_sched_domain(dlb, &args);
}

static int create_ldb_queue(
    dlb_domain_hdl_t domain)
{
    dlb_create_ldb_queue_t args = {0};
    args.num_sequence_numbers = 0;
    return dlb_create_ldb_queue(domain, &args);
}

static int create_ldb_port(
    dlb_domain_hdl_t domain,
    int ldb_pool,
    int dir_pool)
{
    dlb_create_port_t args;
    //if (!cap.combined_credits) {
    if (1) {
        args.ldb_credit_pool_id = ldb_pool;
        args.dir_credit_pool_id = dir_pool;
	// dlb_create_port_t updated since DLB 7.7.0
        //args.num_ldb_credits = 32;
        //args.num_dir_credits = 32;
    } else {
        args.credit_pool_id = ldb_pool;
	// dlb_create_port_t updated since DLB 7.7.0
	//args.num_credits = 32;
    }
    args.cq_depth = CQ_DEPTH;
    args.num_ldb_event_state_entries = CQ_DEPTH*2;
#ifdef DLB2
    args.cos_id = DLB_PORT_COS_ID_ANY;
#endif
    return dlb_create_ldb_port(domain, &args);
}

static ngx_int_t
ngx_loadbalance_init(ngx_thread_pool_t *tp, int index)
{
    int domain_id, tx_port_id, rx_port_id;
    int i, j, ret;
    ngx_ldb_t *ldb_p = &tp->loadbal_queue;

    //TODO, set up dlb configue
    ldb_p->num_credit_combined  = 2048;
    ldb_p->num_credit_ldb       = 2048;
    ldb_p->num_credit_dir       = 1024;

    index = 0;
    ldb_p->dev_id = index % DLB_NUM;

    while (index < DLB_NUM) {
        ret = dlb_open(ldb_p->dev_id, &ldb_p->dlb);
        if (ret < 0) {
            ngx_log_error(NGX_LOG_ERR, tp->log, 0, "dlb_open fail %d errno %d", ret, errno);
            goto retry;
        }

        if (dlb_get_dev_capabilities(ldb_p->dlb, &ldb_p->cap)) {
            ngx_log_error(NGX_LOG_ERR, tp->log, 0, "dlb_get_dev_capabilities errno %u, return", errno);
            return NGX_ERROR;
        }

        if (dlb_get_num_resources(ldb_p->dlb, &ldb_p->rsrcs)) {
            ngx_log_error(NGX_LOG_ERR, tp->log, 0, "dlb_get_dev_capabilities errno %u, return", errno);
            return NGX_ERROR;
        }

        //UNORDER need 0 SN
        unsigned int sns_per_queue;
        ret = dlb_get_ldb_sequence_number_allocation(ldb_p->dlb, 0, &sns_per_queue);
        if (ret < 0 ) {
            ngx_log_error(NGX_LOG_ERR, tp->log, 0, "dlb_set_ldb_sequence_number_allocation errno %d, return", ret, errno);
            return NGX_ERROR;
        }

        domain_id = create_sched_domain(ldb_p->dlb, ldb_p);
        if (domain_id < 0) {
            ngx_log_error(NGX_LOG_ERR, tp->log, 0, "create_sched_domain errno %d, index %d ", errno, index);
            goto retry;
        } else {
            //success
            ldb_p->domain_id = domain_id;
            ldb_p->domain = dlb_attach_sched_domain(ldb_p->dlb, domain_id);
            if (ldb_p->domain == NULL) {
                ngx_log_error(NGX_LOG_ERR, tp->log, 0, "dlb_attach_sched_domain errno %d", errno);
                return NGX_ERROR;
            }
            if (!ldb_p->cap.combined_credits) {
                int max_ldb_credits = ldb_p->rsrcs.num_ldb_credits * partial_resources / 100;
                int max_dir_credits = ldb_p->rsrcs.num_dir_credits * partial_resources / 100;

                ldb_p->ldb_pool_id = dlb_create_ldb_credit_pool(ldb_p->domain, max_ldb_credits);
                if (ldb_p->ldb_pool_id == -1) {
                    ngx_log_error(NGX_LOG_ERR, tp->log, 0, "dlb_create_ldb_credit_pool number %d, errno %d", max_ldb_credits, errno);
                    goto retry;
                }
                ldb_p->dir_pool_id = dlb_create_dir_credit_pool(ldb_p->domain, max_dir_credits);
                if (ldb_p->dir_pool_id == -1) {
                    ngx_log_error(NGX_LOG_ERR, tp->log, 0, "dlb_create_dir_credit_pool errno %d", errno);
                    goto retry;
                }
            } else {
                int max_credits = ldb_p->rsrcs.num_credits * partial_resources / 100;

                ldb_p->ldb_pool_id = dlb_create_credit_pool(ldb_p->domain, max_credits);
                if (ldb_p->ldb_pool_id == -1) {
                    ngx_log_error(NGX_LOG_ERR, tp->log, 0, "dlb_create_credit_pool errno %d", errno);
                    goto retry;
                }
            }

            break;
        }

retry:
        ldb_p->dev_id = ++index;
    }

    if (index >= DLB_NUM)  {
        ngx_log_error(NGX_LOG_ERR, tp->log, 0, "no DLB resource left, used up to %d", index);
        return NGX_ERROR;
    }

    // Create QUEUE_NUM unordered ldb queues
    for (i=0; i < QUEUE_NUM; i++) {
        ldb_p->queue_id[i] = create_ldb_queue(ldb_p->domain);
        if (ldb_p->queue_id[i] == -1) {
            ngx_log_error(NGX_LOG_ERR, tp->log, 0, "create_ldb_queue %d errno %d", i, errno);
            return NGX_ERROR;
        }
    }

    tx_port_id = create_ldb_port(ldb_p->domain, ldb_p->ldb_pool_id, ldb_p->dir_pool_id);
    if (tx_port_id == -1) {
        ngx_log_error(NGX_LOG_ERR, tp->log, 0, "dlb_create_credit_pool errno %d", errno);
        return NGX_ERROR;
    }

    ldb_p->tx_port = dlb_attach_ldb_port(ldb_p->domain, tx_port_id);
    if (ldb_p->tx_port == NULL) {
        ngx_log_error(NGX_LOG_ERR, tp->log, 0, "dlb_create_credit_pool errno %d", errno);
        return NGX_ERROR;
    }

    // Create unordered ldb ports
    // TODO, also need check the max rx number
    //for (i = 0; i < (int)tp->threads; i++) {
    for (i = 0; i < 1; i++) {
        rx_port_id = create_ldb_port(ldb_p->domain, ldb_p->ldb_pool_id, ldb_p->dir_pool_id);
        if (rx_port_id == -1) {
            ngx_log_error(NGX_LOG_ERR, tp->log, 0, "dlb_create_credit_pool errno %d", errno);
            return NGX_ERROR;
        }

        ldb_p->rx_port[i] = dlb_attach_ldb_port(ldb_p->domain, rx_port_id);
        if (ldb_p->rx_port[i] == NULL) {
            ngx_log_error(NGX_LOG_ERR, tp->log, 0, "dlb_create_credit_pool errno %d", errno);
            return NGX_ERROR;
        }

        //link the rx port to the dlb queue
        for ( j=0; j < QUEUE_NUM; j++) {
            if (dlb_link_queue(ldb_p->rx_port[i], ldb_p->queue_id[j], j*2) == -1) {
                ngx_log_error(NGX_LOG_ERR, tp->log, 0, "dlb_link_queue %d errno %d", j, errno);
                return NGX_ERROR;
            }
        }
    }

    if (dlb_launch_domain_alert_thread(ldb_p->domain, NULL, NULL)) {
        ngx_log_error(NGX_LOG_ERR, tp->log, 0, "dlb_launch_domain_alert_thread errno %d", errno);
        return NGX_ERROR;
    }

    if (dlb_start_sched_domain(ldb_p->domain)) {
        ngx_log_error(NGX_LOG_ERR, tp->log, 0, "dlb_start_sched_domain errno %d", errno);
        return NGX_ERROR;
    }

    ngx_log_error(NGX_LOG_ERR, tp->log, 0, "DLB created Success on device %d", ldb_p->dev_id);

    return NGX_OK;
}
#else
static ngx_int_t
ngx_loadbalance_init(ngx_thread_pool_t *tp, int index) {
    return NGX_ERROR;
}
#endif

static ngx_int_t
ngx_thread_pool_init(ngx_thread_pool_t *tp, ngx_log_t *log, ngx_pool_t *pool)
{
    int             err;
    pthread_t       tid;
    ngx_uint_t      n;
    pthread_attr_t  attr;
    ngx_thread_param_t *thread_param = NULL;

    if (ngx_notify == NULL) {
        ngx_log_error(NGX_LOG_ALERT, log, 0,
               "the configured event method cannot be used with thread pools");
        return NGX_ERROR;
    }

    ngx_thread_pool_queue_init(&tp->queue);

    if (ngx_thread_mutex_create(&tp->mtx, log) != NGX_OK) {
        return NGX_ERROR;
    }

    if (ngx_thread_cond_create(&tp->cond, log) != NGX_OK) {
        (void) ngx_thread_mutex_destroy(&tp->mtx, log);
        return NGX_ERROR;
    }

    tp->log = log;

    err = pthread_attr_init(&attr);
    if (err) {
        ngx_log_error(NGX_LOG_ALERT, log, err,
                      "pthread_attr_init() failed");
        return NGX_ERROR;
    }

    err = pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    if (err) {
        ngx_log_error(NGX_LOG_ALERT, log, err,
                      "pthread_attr_setdetachstate() failed");
        return NGX_ERROR;
    }

    if (tp->loadbal) {
        for (n = 0; n < 8; n++) {
            ngx_thread_pool_queue_init(&tp->p_queue[n]);
            tp->p_recieved[n] = 0;
            tp->p_send[n] = 0;
        }

        thread_param = (ngx_thread_param_t*)malloc(sizeof(ngx_thread_param_t)*tp->threads);
        //TODO construct thread param
        for (n = 0; n < tp->threads; n++) {
            //thread_param[n].rx_port = tp->loadbal_queue.rx_port[n];
            thread_param[n].rx_port = tp->loadbal_queue.rx_port[0];
            thread_param[n].tp = tp;
        }
    }

    for (n = 0; n < tp->threads; n++) {
        if (tp->loadbal) {
            err = pthread_create(&tid, &attr, ngx_thread_pool_cycle_loadbalance, &thread_param[n]);
        } else {
            err = pthread_create(&tid, &attr, ngx_thread_pool_cycle, tp);
        }

        if (err) {
            ngx_log_error(NGX_LOG_ALERT, log, err,
                          "pthread_create() failed");
            return NGX_ERROR;
        }
    }

    (void) pthread_attr_destroy(&attr);

    return NGX_OK;
}


static void
ngx_thread_pool_destroy(ngx_thread_pool_t *tp)
{
    ngx_uint_t           n;
    ngx_thread_task_t    task;
    volatile ngx_uint_t  lock;

    ngx_memzero(&task, sizeof(ngx_thread_task_t));

    task.handler = ngx_thread_pool_exit_handler;
    task.ctx = (void *) &lock;

    for (n = 0; n < tp->threads; n++) {
        lock = 1;

        if (ngx_thread_task_post(tp, &task) != NGX_OK) {
            return;
        }

        while (lock) {
            ngx_sched_yield();
        }

        task.event.active = 0;
    }

    if (tp->loadbal && dlb_disable_port(tp->loadbal_queue.rx_port[0]))
        ngx_log_error(NGX_LOG_ERR, tp->log, 0,
                    "dlb_disable_port() rx_port failed with errno %d", errno);

    if (tp->loadbal && dlb_detach_port(tp->loadbal_queue.rx_port[0]) == -1)
        ngx_log_error(NGX_LOG_ERR, tp->log, 0,
                    "dlb_detach_port() rx_port failed with errno %d", errno);

    if (tp->loadbal && dlb_detach_port(tp->loadbal_queue.tx_port) == -1)
        ngx_log_error(NGX_LOG_ERR, tp->log, 0,
                    "dlb_detach_port() tx_port failed with errno %d", errno);

    if (tp->loadbal && dlb_detach_sched_domain(tp->loadbal_queue.domain) == -1)
        ngx_log_error(NGX_LOG_ERR, tp->log, 0,
                        "dlb_detach_sched_domain() failed with errno %d", errno);

    if (tp->loadbal && dlb_reset_sched_domain(tp->loadbal_queue.dlb, tp->loadbal_queue.domain_id) == -1)
        ngx_log_error(NGX_LOG_ERR, tp->log, 0,
                        "dlb_reset_sched_domain() failed with errno %d", errno);

    if (tp->loadbal && dlb_close(tp->loadbal_queue.dlb) == -1)
        ngx_log_error(NGX_LOG_ERR, tp->log, 0,
                        "dlb_close() failed with errno %d", errno);

    (void) ngx_thread_cond_destroy(&tp->cond, tp->log);

    (void) ngx_thread_mutex_destroy(&tp->mtx, tp->log);
}


static void
ngx_thread_pool_exit_handler(void *data, ngx_log_t *log)
{
    ngx_uint_t *lock = data;

    *lock = 0;

    pthread_exit(0);
}


ngx_thread_task_t *
ngx_thread_task_alloc(ngx_pool_t *pool, size_t size)
{
    ngx_thread_task_t  *task;

    task = ngx_pcalloc(pool, sizeof(ngx_thread_task_t) + size);
    if (task == NULL) {
        return NULL;
    }

    task->ctx = task + 1;

    return task;
}

#if (AIO_LOADBALANCE)
ngx_int_t
ngx_thread_task_post_internal(ngx_thread_pool_t *tp, ngx_thread_task_t *task)
{
    int ret;
    dlb_event_t dlb_events[NUM_EVENTS_PER_BATCH];
    static uint16_t i = 0;

    task->event.active = 1;
    task->id = ngx_thread_pool_task_id++;
    task->next = NULL;

    //simply processing, put the task in the sw queue first
    *(tp->p_queue[task->priority].last) = task;
    tp->p_queue[task->priority].last = &task->next;
    tp->p_recieved[task->priority] ++;

    //send the task by priority
    for (int j = 0; j < 8; j++) {
        while(tp->p_queue[j].first != NULL) {
            ngx_thread_task_t *stored_task = tp->p_queue[j].first;
            ngx_thread_task_t **last = tp->p_queue[j].last;
            tp->p_queue[j].first = stored_task->next;
            if (tp->p_queue[j].first == NULL) {
                tp->p_queue[j].last = &tp->p_queue[j].first;
            }

            dlb_events[0].send.queue_id = tp->loadbal_queue.queue_id[stored_task->priority%QUEUE_NUM];
            dlb_events[0].send.sched_type = SCHED_UNORDERED;
            dlb_events[0].send.priority = stored_task->priority * 2;    //priority map to 0/2/4/6
            dlb_events[0].send.udata64 = (uint64_t)stored_task;
            dlb_events[0].send.udata16 = (uint16_t)++i;
            ret = dlb_send(tp->loadbal_queue.tx_port, 1, &dlb_events[0]);

            if (ret <= 0) {
                //restore the queue
                tp->p_queue[j].first = stored_task;
                tp->p_queue[j].last = last;

                if(stored_task->priority != j)
                    ngx_log_error(NGX_LOG_ALERT, tp->log, 0, "send task %d (p:%d, %d) failed with ret %d errno %d, waiting task:%d(%d, %d), %d(%d, %d), %d(%d, %d), %d(%d, %d)",
                        stored_task->id, stored_task->priority, j, ret, errno, tp->p_recieved[0] - tp->p_send[0], tp->p_recieved[0], tp->p_send[0],
                        tp->p_recieved[1] - tp->p_send[1], tp->p_recieved[1], tp->p_send[1],
                        tp->p_recieved[2] - tp->p_send[2], tp->p_recieved[2], tp->p_send[2],
                        tp->p_recieved[3] - tp->p_send[3], tp->p_recieved[3], tp->p_send[3]);
                return NGX_OK;
            } else {
                tp->p_send[j] ++;

                if(stored_task->priority != j)
                    ngx_log_error(NGX_LOG_ALERT, tp->log, 0, "send task %d (p:%d, %d) succeed with ret %d errno %d, waiting task:%d(%d, %d), %d(%d, %d), %d(%d, %d), %d(%d, %d)",
                        stored_task->id, stored_task->priority, j, ret, errno, tp->p_recieved[0] - tp->p_send[0], tp->p_recieved[0], tp->p_send[0],
                        tp->p_recieved[1] - tp->p_send[1], tp->p_recieved[1], tp->p_send[1],
                        tp->p_recieved[2] - tp->p_send[2], tp->p_recieved[2], tp->p_send[2],
                        tp->p_recieved[3] - tp->p_send[3], tp->p_recieved[3], tp->p_send[3]);
            }
        }
    }

    return NGX_OK;
}
#else
ngx_int_t
ngx_thread_task_post_internal(ngx_thread_pool_t *tp, ngx_thread_task_t *task)
{
    return NGX_ERROR;
}
#endif

ngx_int_t
ngx_thread_task_post(ngx_thread_pool_t *tp, ngx_thread_task_t *task)
{
    if (tp->loadbal && ngx_thread_task_post_internal(tp, task) == NGX_OK) {
        return NGX_OK;
    }

    if (task->event.active) {
        ngx_log_error(NGX_LOG_ALERT, tp->log, 0,
                      "task #%ui already active", task->id);
        return NGX_ERROR;
    }

    if (ngx_thread_mutex_lock(&tp->mtx, tp->log) != NGX_OK) {
        return NGX_ERROR;
    }

    if (tp->waiting >= tp->max_queue) {
        (void) ngx_thread_mutex_unlock(&tp->mtx, tp->log);

        ngx_log_error(NGX_LOG_ERR, tp->log, 0,
                      "thread pool \"%V\" queue overflow: %i tasks waiting",
                      &tp->name, tp->waiting);
        return NGX_ERROR;
    }

    task->event.active = 1;

    task->id = ngx_thread_pool_task_id++;
    task->next = NULL;

    if (ngx_thread_cond_signal(&tp->cond, tp->log) != NGX_OK) {
        (void) ngx_thread_mutex_unlock(&tp->mtx, tp->log);
        return NGX_ERROR;
    }

    *tp->queue.last = task;
    tp->queue.last = &task->next;

    tp->waiting++;

    (void) ngx_thread_mutex_unlock(&tp->mtx, tp->log);

    ngx_log_debug2(NGX_LOG_DEBUG_CORE, tp->log, 0,
                   "task #%ui added to thread pool \"%V\"",
                   task->id, &tp->name);

    return NGX_OK;
}

static void *
ngx_thread_pool_cycle_loadbalance(void *data)
{
    ngx_thread_param_t* thread_param = (ngx_thread_param_t*)data;
    ngx_thread_task_t  *task;
    sigset_t            set;
    int ret;
    static uint16_t seq=0;
    //uint16_t send_seq;

    dlb_port_hdl_t rx_port = thread_param->rx_port;
    ngx_thread_pool_t* tp = thread_param->tp;

    dlb_event_t events[NUM_EVENTS_PER_BATCH];

    ngx_log_debug1(NGX_LOG_DEBUG_CORE, tp->log, 0,
                   "ldb thread in pool \"%V\" started", &tp->name);

    sigfillset(&set);

    sigdelset(&set, SIGILL);
    sigdelset(&set, SIGFPE);
    sigdelset(&set, SIGSEGV);
    sigdelset(&set, SIGBUS);

    ret = pthread_sigmask(SIG_BLOCK, &set, NULL);
    if (ret) {
        ngx_log_error(NGX_LOG_ALERT, tp->log, ret, "ldb pthread_sigmask() failed");
        return NULL;
    }

    for ( ;; ) {

        if (ngx_thread_mutex_lock(&tp->mtx, tp->log) != NGX_OK) {
            return NULL;
        }

        ret = dlb_recv(rx_port,
                    NUM_EVENTS_PER_BATCH,
                    true,      //INTERRUPT),
                     events);    //&events[0]);

        /* The port was disabled, indicating the thread should return */
        if (ret == -1 && errno == EACCES){

            ngx_log_error(NGX_LOG_ALERT, tp->log, ret, "dlb_recv() error");
            exit(1);
        }
        // ngx_log_error(NGX_LOG_ALERT, tp->log, ret, "dlb_recv() get one event");
/*
        struct timeval  tv;
        //time_t           sec;
        ngx_uint_t       usec;

        ngx_gettimeofday(&tv);

        //sec = tv.tv_sec;
        usec = tv.tv_usec ;
*/
        if (dlb_release(rx_port, ret) != ret) {
              ngx_log_error(NGX_LOG_ALERT, tp->log, ret, "Failed to release all %d events !\n");
              //exit(-1);
              return NULL;
        }

        if (ngx_thread_mutex_unlock(&tp->mtx, tp->log) != NGX_OK) {
            return NULL;
        }

        for (int i =0; i < ret; i++){
          task = (ngx_thread_task_t*)events[i].recv.udata64;
          //send_seq = (ngx_thread_task_t*)events[i].recv.udata16;

          ngx_log_debug2(NGX_LOG_DEBUG_CORE, tp->log, 0,
                         "run task #%ui in ldb thread pool \"%V\"",
                         task->id, &tp->name);

          task->handler(task->ctx, tp->log);

          seq++;

          //ngx_log_error(NGX_LOG_ALERT, tp->log, 0,
          //          "DLB receive tast %ui priority %i at sequenceR %i send_seq %i at timestampR %d",task->id, task->priority, seq, send_seq, usec);

          ngx_log_debug2(NGX_LOG_DEBUG_CORE, tp->log, 0,
                         "complete task #%ui in ldb thread pool \"%V\"",
                         task->id, &tp->name);

          task->next = NULL;

          ngx_spinlock(&ngx_thread_pool_done_lock, 1, 2048);

          *ngx_thread_pool_done.last = task;
          ngx_thread_pool_done.last = &task->next;

          ngx_memory_barrier();

          ngx_unlock(&ngx_thread_pool_done_lock);

          (void) ngx_notify(ngx_thread_pool_handler);

        }

#if 0
        //sleep 1ms
        for(int tt=0; tt< 1; tt++) {
          /*ngx_gettimeofday(&tv);
          if(tv.tv_usec - usec > 100) {
            ngx_log_error(NGX_LOG_ALERT, tp->log, 0,
                    "DLB receive tast %ui priority %i delay %d us",task->id, task->priority, tv.tv_usec-usec);

            break;
          }
          else */
            //usleep(1000);
        }
#endif

     }
}

static void *
ngx_thread_pool_cycle(void *data)
{
    ngx_thread_pool_t *tp = data;

    int                 err;
    sigset_t            set;
    ngx_thread_task_t  *task;

#if 0
    ngx_time_update();
#endif

    ngx_log_debug1(NGX_LOG_DEBUG_CORE, tp->log, 0,
                   "thread in pool \"%V\" started", &tp->name);

    sigfillset(&set);

    sigdelset(&set, SIGILL);
    sigdelset(&set, SIGFPE);
    sigdelset(&set, SIGSEGV);
    sigdelset(&set, SIGBUS);

    err = pthread_sigmask(SIG_BLOCK, &set, NULL);
    if (err) {
        ngx_log_error(NGX_LOG_ALERT, tp->log, err, "pthread_sigmask() failed");
        return NULL;
    }

    for ( ;; ) {
        if (ngx_thread_mutex_lock(&tp->mtx, tp->log) != NGX_OK) {
            return NULL;
        }

        /* the number may become negative */
        tp->waiting--;

        while (tp->queue.first == NULL) {
            if (ngx_thread_cond_wait(&tp->cond, &tp->mtx, tp->log)
                != NGX_OK)
            {
                (void) ngx_thread_mutex_unlock(&tp->mtx, tp->log);
                return NULL;
            }
        }

        task = tp->queue.first;
        tp->queue.first = task->next;

        if (tp->queue.first == NULL) {
            tp->queue.last = &tp->queue.first;
        }

        if (ngx_thread_mutex_unlock(&tp->mtx, tp->log) != NGX_OK) {
            return NULL;
        }

#if 0
        ngx_time_update();
#endif

        ngx_log_debug2(NGX_LOG_DEBUG_CORE, tp->log, 0,
                       "run task #%ui in thread pool \"%V\"",
                       task->id, &tp->name);

        task->handler(task->ctx, tp->log);

        ngx_log_debug2(NGX_LOG_DEBUG_CORE, tp->log, 0,
                       "complete task #%ui in thread pool \"%V\"",
                       task->id, &tp->name);

        task->next = NULL;

        ngx_spinlock(&ngx_thread_pool_done_lock, 1, 2048);

        *ngx_thread_pool_done.last = task;
        ngx_thread_pool_done.last = &task->next;

        ngx_memory_barrier();

        ngx_unlock(&ngx_thread_pool_done_lock);

        (void) ngx_notify(ngx_thread_pool_handler);
    }
}

static void
ngx_thread_pool_handler(ngx_event_t *ev)
{
    ngx_event_t        *event;
    ngx_thread_task_t  *task;

    ngx_log_debug0(NGX_LOG_DEBUG_CORE, ev->log, 0, "thread pool handler");

    ngx_spinlock(&ngx_thread_pool_done_lock, 1, 2048);

    task = ngx_thread_pool_done.first;
    ngx_thread_pool_done.first = NULL;
    ngx_thread_pool_done.last = &ngx_thread_pool_done.first;

    ngx_memory_barrier();

    ngx_unlock(&ngx_thread_pool_done_lock);

    while (task) {
        ngx_log_debug1(NGX_LOG_DEBUG_CORE, ev->log, 0,
                       "run completion handler for task #%ui", task->id);

        event = &task->event;
        task = task->next;

        event->complete = 1;
        event->active = 0;

        event->handler(event);
    }
}


static void *
ngx_thread_pool_create_conf(ngx_cycle_t *cycle)
{
    ngx_thread_pool_conf_t  *tcf;

    tcf = ngx_pcalloc(cycle->pool, sizeof(ngx_thread_pool_conf_t));
    if (tcf == NULL) {
        return NULL;
    }

    if (ngx_array_init(&tcf->pools, cycle->pool, 4,
                       sizeof(ngx_thread_pool_t *))
        != NGX_OK)
    {
        return NULL;
    }

    return tcf;
}


static char *
ngx_thread_pool_init_conf(ngx_cycle_t *cycle, void *conf)
{
    ngx_thread_pool_conf_t *tcf = conf;

    ngx_uint_t           i;
    ngx_thread_pool_t  **tpp;

    tpp = tcf->pools.elts;

    for (i = 0; i < tcf->pools.nelts; i++) {

        if (tpp[i]->threads) {
            continue;
        }

        if (tpp[i]->name.len == ngx_thread_pool_default.len
            && ngx_strncmp(tpp[i]->name.data, ngx_thread_pool_default.data,
                           ngx_thread_pool_default.len)
               == 0)
        {
            tpp[i]->threads = 32;
            tpp[i]->max_queue = 65536;
            continue;
        }

        ngx_log_error(NGX_LOG_EMERG, cycle->log, 0,
                      "unknown thread pool \"%V\" in %s:%ui",
                      &tpp[i]->name, tpp[i]->file, tpp[i]->line);

        return NGX_CONF_ERROR;
    }

    return NGX_CONF_OK;
}


static char *
ngx_thread_pool(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    ngx_str_t          *value;
    ngx_uint_t          i;
    ngx_thread_pool_t  *tp;

    value = cf->args->elts;

    tp = ngx_thread_pool_add(cf, &value[1]);

    if (tp == NULL) {
        return NGX_CONF_ERROR;
    }

    if (tp->threads) {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "duplicate thread pool \"%V\"", &tp->name);
        return NGX_CONF_ERROR;
    }

    tp->max_queue = 65536;

    for (i = 2; i < cf->args->nelts; i++) {

        if (ngx_strncmp(value[i].data, "threads=", 8) == 0) {

            tp->threads = ngx_atoi(value[i].data + 8, value[i].len - 8);

            if (tp->threads == (ngx_uint_t) NGX_ERROR || tp->threads == 0) {
                ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                                   "invalid threads value \"%V\"", &value[i]);
                return NGX_CONF_ERROR;
            }

            continue;
        }

        if (ngx_strncmp(value[i].data, "max_queue=", 10) == 0) {

            tp->max_queue = ngx_atoi(value[i].data + 10, value[i].len - 10);

            if (tp->max_queue == NGX_ERROR) {
                ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                                   "invalid max_queue value \"%V\"", &value[i]);
                return NGX_CONF_ERROR;
            }

            continue;
        }
    }

    if (tp->threads == 0) {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "\"%V\" must have \"threads\" parameter",
                           &cmd->name);
        return NGX_CONF_ERROR;
    }

    return NGX_CONF_OK;
}


ngx_thread_pool_t *
ngx_thread_pool_add(ngx_conf_t *cf, ngx_str_t *name)
{
    ngx_thread_pool_t       *tp, **tpp;
    ngx_thread_pool_conf_t  *tcf;

    if (name == NULL) {
        name = &ngx_thread_pool_default;
    }

    tp = ngx_thread_pool_get(cf->cycle, name);

    if (tp) {
        return tp;
    }

    tp = ngx_pcalloc(cf->pool, sizeof(ngx_thread_pool_t));
    if (tp == NULL) {
        return NULL;
    }

    tp->name = *name;
    tp->file = cf->conf_file->file.name.data;
    tp->line = cf->conf_file->line;

    tcf = (ngx_thread_pool_conf_t *) ngx_get_conf(cf->cycle->conf_ctx,
                                                  ngx_thread_pool_module);

    tpp = ngx_array_push(&tcf->pools);
    if (tpp == NULL) {
        return NULL;
    }

    *tpp = tp;

    return tp;
}


ngx_thread_pool_t *
ngx_thread_pool_get(ngx_cycle_t *cycle, ngx_str_t *name)
{
    ngx_uint_t                i;
    ngx_thread_pool_t       **tpp;
    ngx_thread_pool_conf_t   *tcf;

    tcf = (ngx_thread_pool_conf_t *) ngx_get_conf(cycle->conf_ctx,
                                                  ngx_thread_pool_module);

    tpp = tcf->pools.elts;

    for (i = 0; i < tcf->pools.nelts; i++) {

        if (tpp[i]->name.len == name->len
            && ngx_strncmp(tpp[i]->name.data, name->data, name->len) == 0)
        {
            return tpp[i];
        }
    }

    return NULL;
}


static ngx_int_t
ngx_thread_pool_init_worker(ngx_cycle_t *cycle)
{
    ngx_uint_t                i;
    ngx_thread_pool_t       **tpp;
    ngx_thread_pool_conf_t   *tcf;

    if (ngx_process != NGX_PROCESS_WORKER
        && ngx_process != NGX_PROCESS_SINGLE
        && !ngx_is_privileged_agent)
    {
        return NGX_OK;
    }

    tcf = (ngx_thread_pool_conf_t *) ngx_get_conf(cycle->conf_ctx,
                                                  ngx_thread_pool_module);

    if (tcf == NULL) {
        return NGX_OK;
    }

    ngx_thread_pool_queue_init(&ngx_thread_pool_done);

    tpp = tcf->pools.elts;

    for (i = 0; i < tcf->pools.nelts; i++) {
        tpp[i]->log = cycle->log;

        if (ngx_loadbalance_init(tpp[i], i) == NGX_OK) {
            tpp[i]->loadbal = 1;
        } else {
            ngx_log_error(NGX_LOG_ERR, tpp[i]->log, 0, "dlb load balance init fail, errno %d", errno);
            // fallback to no loadbalance
            tpp[i]->loadbal = 0;
        }

        if (ngx_thread_pool_init(tpp[i], cycle->log, cycle->pool) != NGX_OK) {
            return NGX_ERROR;
        }
    }

    return NGX_OK;
}


static void
ngx_thread_pool_exit_worker(ngx_cycle_t *cycle)
{
    ngx_uint_t                i;
    ngx_thread_pool_t       **tpp;
    ngx_thread_pool_conf_t   *tcf;

    if (ngx_process != NGX_PROCESS_WORKER
        && ngx_process != NGX_PROCESS_SINGLE)
    {
        return;
    }

    tcf = (ngx_thread_pool_conf_t *) ngx_get_conf(cycle->conf_ctx,
                                                  ngx_thread_pool_module);

    if (tcf == NULL) {
        return;
    }

    tpp = tcf->pools.elts;

    for (i = 0; i < tcf->pools.nelts; i++) {
        ngx_thread_pool_destroy(tpp[i]);
    }
}
